# sage_setup: distribution = sagemath-combinat
# sage.doctest: needs sage.combinat sage.modules
"""
Free algebra quotient elements

AUTHORS:
    - William Stein (2011-11-19): improved doctest coverage to 100%
    - David Kohel (2005-09): initial version
"""

#*****************************************************************************
#  Copyright (C) 2005 David Kohel <kohel@maths.usyd.edu>
#
#  Distributed under the terms of the GNU General Public License (GPL)
#
#    This code is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty
#    of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
#
#  See the GNU General Public License for more details; the full text
#  is available at:
#
#                  http://www.gnu.org/licenses/
#*****************************************************************************

from sage.misc.repr import repr_lincomb
from sage.structure.element import RingElement, AlgebraElement
from sage.structure.parent_gens import localvars
from sage.structure.richcmp import richcmp
from sage.rings.integer import Integer
from sage.modules.free_module_element import FreeModuleElement
from sage.monoids.free_monoid_element import FreeMonoidElement
from sage.algebras.free_algebra_element import FreeAlgebraElement


def is_FreeAlgebraQuotientElement(x):
    """
    EXAMPLES::

        sage: from sage.algebras.free_algebra_quotient import hamilton_quatalg
        sage: from sage.algebras.free_algebra_quotient_element import is_FreeAlgebraQuotientElement
        sage: H, (i,j,k) = hamilton_quatalg(QQ)
        sage: is_FreeAlgebraQuotientElement(i)
        doctest:warning...
        DeprecationWarning: The function is_FreeAlgebraQuotientElement is deprecated;
        use 'isinstance(..., FreeAlgebraQuotientElement)' instead.
        See https://github.com/sagemath/sage/issues/38184 for details.
        True

    Of course this is testing the data type::

        sage: is_FreeAlgebraQuotientElement(1)
        False
        sage: is_FreeAlgebraQuotientElement(H(1))
        True
    """
    from sage.misc.superseded import deprecation
    deprecation(38184,
                "The function is_FreeAlgebraQuotientElement is deprecated; "
                "use 'isinstance(..., FreeAlgebraQuotientElement)' instead.")
    return isinstance(x, FreeAlgebraQuotientElement)


class FreeAlgebraQuotientElement(AlgebraElement):
    def __init__(self, A, x):
        """
        Create the element x of the FreeAlgebraQuotient A.

        EXAMPLES::

            sage: from sage.algebras.free_algebra_quotient import FreeAlgebraQuotientElement
            sage: H, (i,j,k) = sage.algebras.free_algebra_quotient.hamilton_quatalg(ZZ)
            sage: FreeAlgebraQuotientElement(H, i)
            i
            sage: a = FreeAlgebraQuotientElement(H, 1); a
            1
            sage: a in H
            True

        TESTS::

            sage: TestSuite(i).run()
        """
        AlgebraElement.__init__(self, A)
        Q = self.parent()

        if isinstance(x, FreeAlgebraQuotientElement) and x.parent() == Q:
            self.__vector = Q.module()(x.vector())
            return
        if isinstance(x, (Integer, int)):
            self.__vector = Q.module().gen(0) * x
            return
        elif isinstance(x, FreeModuleElement) and x.parent() is Q.module():
            self.__vector = x
            return
        elif isinstance(x, FreeModuleElement) and x.parent() == A.module():
            self.__vector = x
            return
        R = A.base_ring()
        M = A.module()
        F = A.monoid()
        B = A.monomial_basis()

        if isinstance(x, (Integer, int)):
            self.__vector = x*M.gen(0)
        elif isinstance(x, RingElement) and not isinstance(x, AlgebraElement) and x in R:
            self.__vector = x * M.gen(0)
        elif isinstance(x, FreeMonoidElement) and x.parent() is F:
            if x in B:
                self.__vector = M.gen(B.index(x))
            else:
                raise AttributeError("argument x (= %s) is not in monomial basis" % x)
        elif isinstance(x, list) and len(x) == A.dimension():
            try:
                self.__vector = M(x)
            except TypeError:
                raise TypeError("argument x (= %s) is of the wrong type" % x)
        elif isinstance(x, FreeAlgebraElement) and x.parent() is A.free_algebra():
            # Need to do more work here to include monomials not
            # represented in the monomial basis.
            self.__vector = M(0)
            for m, c in x._FreeAlgebraElement__monomial_coefficients.items():
                self.__vector += c*M.gen(B.index(m))
        elif isinstance(x, dict):
            self.__vector = M(0)
            for m, c in x.items():
                self.__vector += c*M.gen(B.index(m))
        elif isinstance(x, AlgebraElement) and x.parent().ambient_algebra() is A:
            self.__vector = x.ambient_algebra_element().vector()
        else:
            raise TypeError("argument x (= %s) is of the wrong type" % x)

    def _repr_(self):
        """
        EXAMPLES::

            sage: H, (i,j,k) = sage.algebras.free_algebra_quotient.hamilton_quatalg(ZZ)
            sage: i._repr_()
            'i'
        """
        Q = self.parent()
        M = Q.monoid()
        with localvars(M, Q.variable_names()):
            cffs = list(self.__vector)
            mons = Q.monomial_basis()
            return repr_lincomb(zip(mons, cffs), strip_one=True)

    def _latex_(self):
        r"""
        EXAMPLES::

            sage: H, (i,j,k) = sage.algebras.free_algebra_quotient.hamilton_quatalg(QQ)
            sage: ((2/3)*i - j)._latex_()
            '\\frac{2}{3} i - j'
        """
        Q = self.parent()
        M = Q.monoid()
        with localvars(M, Q.variable_names()):
            cffs = tuple(self.__vector)
            mons = Q.monomial_basis()
            return repr_lincomb(zip(mons, cffs), is_latex=True, strip_one=True)

    def vector(self):
        """
        Return underlying vector representation of this element.

        EXAMPLES::

            sage: H, (i,j,k) = sage.algebras.free_algebra_quotient.hamilton_quatalg(QQ)
            sage: ((2/3)*i - j).vector()
            (0, 2/3, -1, 0)
        """
        return self.__vector

    def _richcmp_(self, right, op):
        """
        Compare two quotient algebra elements; done by comparing the
        underlying vector representatives.

        EXAMPLES::

            sage: H, (i,j,k) = sage.algebras.free_algebra_quotient.hamilton_quatalg(QQ)
            sage: i > j
            True
            sage: i == i
            True
            sage: i == 1
            False
            sage: i + j == j + i
            True
        """
        return richcmp(self.vector(), right.vector(), op)

    def __neg__(self):
        """
        Return negative of ``self``.

        EXAMPLES::

            sage: H, (i,j,k) = sage.algebras.free_algebra_quotient.hamilton_quatalg(QQ)
            sage: -i
            -i
            sage: -(2/3*i - 3/7*j + k)
            -2/3*i + 3/7*j - k
        """
        y = self.parent()(0)
        y.__vector = -self.__vector
        return y

    def _add_(self, y):
        """
        EXAMPLES::

            sage: H, (i,j,k) = sage.algebras.free_algebra_quotient.hamilton_quatalg(QQ)
            sage: 2/3*i + 4*j + k
            2/3*i + 4*j + k
        """
        A = self.parent()
        z = A(0)
        z.__vector = self.__vector + y.__vector
        return z

    def _sub_(self, y):
        """
        EXAMPLES::

            sage: H, (i,j,k) = sage.algebras.free_algebra_quotient.hamilton_quatalg(QQ)
            sage: 2/3*i - 4*j
            2/3*i - 4*j
            sage: a = 2/3*i - 4*j; a
            2/3*i - 4*j
            sage: a - a
            0
        """
        A = self.parent()
        z = A(0)
        z.__vector = self.__vector - y.__vector
        return z

    def _mul_(self, y):
        """
        EXAMPLES::

            sage: H, (i,j,k) = sage.algebras.free_algebra_quotient.hamilton_quatalg(QQ)
            sage: a = (5 + 2*i - 3/5*j + 17*k); a*(a+10)
            -5459/25 + 40*i - 12*j + 340*k

        Double check that the above is actually right::

            sage: R.<i,j,k> = QuaternionAlgebra(QQ,-1,-1)                               # needs sage.libs.singular
            sage: a = (5 + 2*i - 3/5*j + 17*k); a*(a+10)                                # needs sage.libs.singular
            -5459/25 + 40*i - 12*j + 340*k
        """
        A = self.parent()

        def monomial_product(X, w, m):
            mats = X._FreeAlgebraQuotient__matrix_action
            for j, k in m._element_list:
                M = mats[int(j)]
                for _ in range(k):
                    w *= M
            return w
        u = self.__vector.__copy__()
        v = y.__vector
        z = A(0)
        B = A.monomial_basis()
        for i in range(A.dimension()):
            c = v[i]
            if c != 0:
                z.__vector += monomial_product(A,c*u,B[i])
        return z

    def _rmul_(self, c):
        """
        EXAMPLES::

            sage: H, (i,j,k) = sage.algebras.free_algebra_quotient.hamilton_quatalg(QQ)
            sage: 3 * (-1+i-2*j+k)
            -3 + 3*i - 6*j + 3*k
            sage: (-1+i-2*j+k)._rmul_(3)
            -3 + 3*i - 6*j + 3*k
        """
        return self.parent([c*a for a in self.__vector])

    def _lmul_(self, c):
        """
        EXAMPLES::

            sage: H, (i,j,k) = sage.algebras.free_algebra_quotient.hamilton_quatalg(QQ)
            sage: (-1+i-2*j+k) * 3
            -3 + 3*i - 6*j + 3*k
            sage: (-1+i-2*j+k)._lmul_(3)
            -3 + 3*i - 6*j + 3*k
        """
        return self.parent([a*c for a in self.__vector])



# sage_setup: distribution = sagemath-combinat
"""
Elements of Free Monoids

AUTHORS:

- David Kohel (2005-09-29)

Elements of free monoids are represented internally as lists of
pairs of integers.
"""

# ****************************************************************************
#  Copyright (C) 2005 David Kohel <kohel@maths.usyd.edu>
#
#  Distributed under the terms of the GNU General Public License (GPL)
#
#    This code is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty
#    of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
#
#  See the GNU General Public License for more details; the full text
#  is available at:
#
#                  https://www.gnu.org/licenses/
# ****************************************************************************

from sage.rings.integer import Integer
from sage.structure.element import MonoidElement
from sage.structure.richcmp import richcmp, richcmp_not_equal
from sage.rings.semirings.non_negative_integer_semiring import NN


def is_FreeMonoidElement(x):
    from sage.misc.superseded import deprecation
    deprecation(38184,
                "The function is_FreeMonoidElement is deprecated; "
                "use 'isinstance(..., FreeMonoidElement)' instead.")
    return isinstance(x, FreeMonoidElement)


class FreeMonoidElement(MonoidElement):
    """
    Element of a free monoid.

    EXAMPLES::

        sage: a = FreeMonoid(5, 'a').gens()
        sage: x = a[0]*a[1]*a[4]**3
        sage: x**3
        a0*a1*a4^3*a0*a1*a4^3*a0*a1*a4^3
        sage: x**0
        1
        sage: x**(-1)
        Traceback (most recent call last):
        ...
        NotImplementedError
    """
    def __init__(self, F, x, check=True):
        """
        Create the element `x` of the FreeMonoid `F`.

        This should typically be called by a FreeMonoid.
        """
        MonoidElement.__init__(self, F)
        if isinstance(x, (int, Integer)):
            if x == 1:
                self._element_list = []
            else:
                raise TypeError("argument x (= %s) is of the wrong type" % x)
        elif isinstance(x, list):
            if check:
                x2 = []
                for v in x:
                    if not (isinstance(v, tuple) and len(v) == 2):
                        raise TypeError("x (= %s) must be a list of 2-tuples or 1" % x)
                    if not (isinstance(v[0], (int, Integer)) and
                            isinstance(v[1], (int, Integer))):
                        raise TypeError("x (= %s) must be a list of 2-tuples of integers or 1" % x)
                    if len(x2) > 0 and v[0] == x2[len(x2)-1][0]:
                        x2[len(x2)-1] = (v[0], v[1]+x2[len(x2)-1][1])
                    else:
                        x2.append(v)
                self._element_list = x2
            else:
                self._element_list = list(x)  # make copy, so user can't accidentally change monoid.

        else:
            # TODO: should have some other checks here...
            raise TypeError("argument x (= %s) is of the wrong type" % x)

    def __hash__(self):
        r"""
        TESTS::

            sage: R.<x,y> = FreeMonoid(2)
            sage: hash(x) == hash(((0, 1),))
            True
            sage: hash(y) == hash(((1, 1),))
            True
            sage: hash(x*y) == hash(((0, 1), (1, 1)))
            True
        """
        return hash(tuple(self._element_list))

    def __iter__(self):
        """
        Return an iterator which yields tuples of variable and exponent.

        EXAMPLES::

            sage: a = FreeMonoid(5, 'a').gens()
            sage: list(a[0]*a[1]*a[4]**3*a[0])
            [(a0, 1), (a1, 1), (a4, 3), (a0, 1)]
        """
        gens = self.parent().gens()
        return ((gens[index], exponent)
                for (index, exponent) in self._element_list)

    def _repr_(self):
        s = ""
        v = self._element_list
        x = self.parent().variable_names()
        for i in range(len(v)):
            if len(s) > 0:
                s += "*"
            g = x[int(v[i][0])]
            e = v[i][1]
            if e == 1:
                s += "%s" % g
            else:
                s += f"{g}^{e}"
        if len(s) == 0:
            s = "1"
        return s

    def _latex_(self) -> str:
        r"""
        Return latex representation of ``self``.

        EXAMPLES::

            sage: F = FreeMonoid(3, 'a')
            sage: z = F([(0,5),(1,2),(0,10),(0,2),(1,2)])
            sage: z._latex_()
            'a_{0}^{5}a_{1}^{2}a_{0}^{12}a_{1}^{2}'
            sage: F.<alpha,beta,gamma> = FreeMonoid(3)
            sage: latex(alpha*beta*gamma)
            \alpha \beta \gamma

        Check that :issue:`14509` is fixed::

            sage: # needs sage.symbolic
            sage: K.< alpha,b > = FreeAlgebra(SR)
            sage: latex(alpha*b)
            \alpha b
            sage: latex(b*alpha)
            b \alpha
            sage: "%s" % latex(alpha*b)
            '\\alpha b'
        """
        s = ""
        v = self._element_list
        x = self.parent().latex_variable_names()
        for i in range(len(v)):
            g = x[int(v[i][0])]
            e = v[i][1]
            if e == 1:
                s += f"{g} "
            else:
                s += f"{g}^{{{e}}}"
        s = s.rstrip(" ")  # strip the trailing whitespace caused by adding a space after each element name
        if len(s) == 0:
            s = "1"
        return s

    def __call__(self, *x, **kwds):
        """
        EXAMPLES::

            sage: M.<x,y> = FreeMonoid(2)
            sage: (x*y).substitute(x=1)
            y

            sage: M.<a> = FreeMonoid(1)
            sage: a.substitute(a=5)
            5

            sage: M.<x,y,z> = FreeMonoid(3)
            sage: (x*y).subs(x=1,y=2,z=14)
            2
            sage: (x*y).subs({x:z,y:z})
            z^2

        It is still possible to substitute elements
        that have no common parent::

            sage: M1 = MatrixSpace(ZZ,1,2)                                              # needs sage.modules
            sage: M2 = MatrixSpace(ZZ,2,1)                                              # needs sage.modules
            sage: (x*y).subs({x: M1([1,2]), y: M2([3,4])})                              # needs sage.modules
            [11]

        TESTS::

            sage: M.<x,y> = FreeMonoid(2)
            sage: (x*y)(QQ(4),QQ(5)).parent()
            Rational Field

        The codomain is by default the first parent::

            sage: M.one()(QQ(4),QQ(5)).parent()
            Rational Field

        unless there is no variable and no substitution::

            sage: M = FreeMonoid(0, [])
            sage: M.one()().parent()
            Free monoid on 0 generators ()

        AUTHORS:

        - Joel B. Mohler (2007-10-27)
        """
        if kwds and x:
            raise ValueError("must not specify both a keyword and positional argument")

        P = self.parent()

        if kwds:
            x = self.gens()
            gens_dict = {name: i for i, name in enumerate(P.variable_names())}
            for key, value in kwds.items():
                if key in gens_dict:
                    x[gens_dict[key]] = value

        if x and isinstance(x[0], tuple):
            x = x[0]

        if len(x) != self.parent().ngens():
            raise ValueError("must specify as many values as generators in parent")

        # if no substitution, do nothing
        if not x:
            return self

        try:
            # This will land in the parent of the first element
            result = x[0].parent().one()
        except (AttributeError, TypeError):
            # unless the parent has no unit
            result = NN.one()
        for var_index, exponent in self._element_list:
            replacement = x[var_index]
            if exponent > 1:
                result *= replacement ** exponent
            elif exponent == 1:
                result *= replacement
        return result

    def _mul_(self, y):
        """
        Multiply two elements ``self`` and ``y`` of the
        free monoid.

        EXAMPLES::

            sage: a = FreeMonoid(5, 'a').gens()
            sage: x = a[0] * a[1] * a[4]**3
            sage: y = a[4] * a[0] * a[1]
            sage: x*y
            a0*a1*a4^4*a0*a1
        """
        M = self.parent()
        z = M(1)
        x_elt = self._element_list
        y_elt = y._element_list
        if not x_elt:
            z._element_list = y_elt
        elif not y_elt:
            z._element_list = x_elt
        else:
            k = len(x_elt)-1
            if x_elt[k][0] != y_elt[0][0]:
                z._element_list = x_elt + y_elt
            else:
                m = (y_elt[0][0], x_elt[k][1]+y_elt[0][1])
                z._element_list = x_elt[:k] + [m] + y_elt[1:]
        return z

    def __invert__(self):
        """
        EXAMPLES::

            sage: a = FreeMonoid(5, 'a').gens()
            sage: x = a[0]*a[1]*a[4]**3
            sage: x**(-1)
            Traceback (most recent call last):
            ...
            NotImplementedError
        """
        raise NotImplementedError

    def __len__(self) -> int:
        """
        Return the degree of the monoid element ``self``, where each
        generator of the free monoid is given degree `1`.

        For example, the length of the identity is `0`, and the
        length of `x_0^2x_1` is `3`.

        EXAMPLES::

            sage: F = FreeMonoid(3, 'a')
            sage: z = F(1)
            sage: len(z)
            0
            sage: a = F.gens()
            sage: len(a[0]**2 * a[1])
            3
        """
        return sum(x[1] for x in self._element_list)

    def _richcmp_(self, other, op) -> bool:
        """
        Compare two free monoid elements with the same parents.

        The ordering is first by increasing length, then lexicographically
        on the underlying word.

        EXAMPLES::

            sage: S = FreeMonoid(3, 'a')
            sage: (x,y,z) = S.gens()
            sage: x * y < y * x
            True

            sage: a = FreeMonoid(5, 'a').gens()
            sage: x = a[0]*a[1]*a[4]**3
            sage: x < x
            False
            sage: x == x
            True
            sage: x >= x*x
            False
        """
        m = sum(i for x, i in self._element_list)
        n = sum(i for x, i in other._element_list)
        if m != n:
            return richcmp_not_equal(m, n, op)
        v = tuple([x for x, i in self._element_list for j in range(i)])
        w = tuple([x for x, i in other._element_list for j in range(i)])
        return richcmp(v, w, op)

    def _acted_upon_(self, x, self_on_left):
        """
        Return the action of the integer 1 on this element.

        EXAMPLES::

            sage: M.<x,y,z>=FreeMonoid(3)
            sage: 1*x
            x
        """
        if x == 1:
            return self
        return None

    def to_word(self, alph=None):
        """
        Return ``self`` as a word.

        INPUT:

        - ``alph`` -- (optional) the alphabet which the result should
          be specified in

        EXAMPLES::

            sage: M.<x,y,z> = FreeMonoid(3)
            sage: a = x * x * y * x
            sage: w = a.to_word(); w
            word: xxyx
            sage: w.to_monoid_element() == a
            True

        .. SEEALSO::

            :meth:`to_list`
        """
        from sage.combinat.words.finite_word import Words
        gens = self.parent().gens()
        if alph is None:
            alph = gens
        alph = [str(c) for c in alph]
        W = Words(alph, infinite=False)
        return W(sum([[alph[gens.index(i[0])]] * i[1] for i in self], []))

    def to_list(self, indices=False) -> list:
        r"""
        Return ``self`` as a list of generators.

        If ``self`` equals `x_{i_1} x_{i_2} \cdots x_{i_n}`, with
        `x_{i_1}, x_{i_2}, \ldots, x_{i_n}` being some of the
        generators of the free monoid, then this method returns
        the list `[x_{i_1}, x_{i_2}, \ldots, x_{i_n}]`.

        If the optional argument ``indices`` is set to ``True``,
        then the list `[i_1, i_2, \ldots, i_n]` is returned instead.

        EXAMPLES::

            sage: M.<x,y,z> = FreeMonoid(3)
            sage: a = x * x * y * x
            sage: w = a.to_list(); w
            [x, x, y, x]
            sage: M.prod(w) == a
            True
            sage: w = a.to_list(indices=True); w
            [0, 0, 1, 0]
            sage: a = M.one()
            sage: a.to_list()
            []

        .. SEEALSO::

            :meth:`to_word`
        """
        if not indices:
            return sum(([i[0]] * i[1] for i in list(self)), [])
        gens = self.parent().gens()
        return sum(([gens.index(i[0])] * i[1] for i in list(self)), [])



# sage_setup: distribution = sagemath-combinat
# sage.doctest: needs sage.combinat sage.graphs sage.modules
r"""
Bijection classes for type `B_n^{(1)}`

Part of the (internal) classes which runs the bijection between rigged
configurations and KR tableaux of type `B_n^{(1)}`.

AUTHORS:

- Travis Scrimshaw (2012-12-21): Initial version

TESTS::

    sage: KRT = crystals.TensorProductOfKirillovReshetikhinTableaux(['B', 3, 1], [[2,1]])
    sage: from sage.combinat.rigged_configurations.bij_type_B import KRTToRCBijectionTypeB
    sage: bijection = KRTToRCBijectionTypeB(KRT(pathlist=[[-1,2]]))
    sage: TestSuite(bijection).run()
    sage: RC = RiggedConfigurations(['B', 3, 1], [[2, 1]])
    sage: from sage.combinat.rigged_configurations.bij_type_B import RCToKRTBijectionTypeB
    sage: bijection = RCToKRTBijectionTypeB(RC(partition_list=[[],[],[]]))
    sage: TestSuite(bijection).run()
"""

# ****************************************************************************
#       Copyright (C) 2012 Travis Scrimshaw <tscrim@ucdavis.edu>
#
#  Distributed under the terms of the GNU General Public License (GPL)
#
#    This code is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
#    General Public License for more details.
#
#  The full text of the GPL is available at:
#
#                  https://www.gnu.org/licenses/
# ****************************************************************************

from sage.combinat.rigged_configurations.bij_type_A import KRTToRCBijectionTypeA
from sage.combinat.rigged_configurations.bij_type_C import KRTToRCBijectionTypeC
from sage.combinat.rigged_configurations.bij_type_C import RCToKRTBijectionTypeC


class KRTToRCBijectionTypeB(KRTToRCBijectionTypeC):
    r"""
    Specific implementation of the bijection from KR tableaux to rigged
    configurations for type `B_n^{(1)}`.
    """

    def run(self, verbose=False):
        """
        Run the bijection from a tensor product of KR tableaux to a rigged
        configuration.

        INPUT:

        - ``tp_krt`` -- a tensor product of KR tableaux

        - ``verbose`` -- (default: ``False``) display each step in the
          bijection

        EXAMPLES::

            sage: from sage.combinat.rigged_configurations.bij_type_B import KRTToRCBijectionTypeB
            sage: KRT = crystals.TensorProductOfKirillovReshetikhinTableaux(['B', 3, 1], [[2, 1]])
            sage: KRTToRCBijectionTypeB(KRT(pathlist=[[0,3]])).run()
            <BLANKLINE>
            0[ ]0
            <BLANKLINE>
            -1[ ]-1
            -1[ ]-1
            <BLANKLINE>
            0[]0
            <BLANKLINE>
            sage: KRT = crystals.TensorProductOfKirillovReshetikhinTableaux(['B', 3, 1], [[3, 1]])
            sage: KRTToRCBijectionTypeB(KRT(pathlist=[[-2,3,1]])).run()
            <BLANKLINE>
            (/)
            <BLANKLINE>
            -1[ ]-1
            <BLANKLINE>
            0[]0
            <BLANKLINE>

        TESTS:

        Check that :issue:`19384` is fixed::

            sage: RC = RiggedConfigurations(['B',3,1], [[3,1],[3,1]])
            sage: RC._test_bijection()
            sage: RC = RiggedConfigurations(['B',3,1], [[1,1],[3,1],[1,1]])
            sage: RC._test_bijection()
        """
        if verbose:
            from sage.combinat.rigged_configurations.tensor_product_kr_tableaux_element \
                import TensorProductOfKirillovReshetikhinTableauxElement

        for cur_crystal in reversed(self.tp_krt):
            r = cur_crystal.parent().r()

            # Check if it is a spinor
            if r == self.n:
                # Perform the spinor bijection by converting to type A_{2n-1}^{(2)}
                #   doing the bijection there and pulling back
                from sage.combinat.rigged_configurations.bij_type_A2_odd import KRTToRCBijectionTypeA2Odd
                from sage.combinat.rigged_configurations.tensor_product_kr_tableaux import TensorProductOfKirillovReshetikhinTableaux
                from sage.combinat.rigged_configurations.rigged_partition import RiggedPartition

                if verbose:
                    print("====================")
                    if len(self.cur_path) == 0:
                        print(repr([]))  # Special case for displaying when the rightmost factor is a spinor
                    else:
                        print(repr(TensorProductOfKirillovReshetikhinTableauxElement(self.tp_krt.parent(), self.cur_path)))
                    print("--------------------")
                    print(repr(self.ret_rig_con))
                    print("--------------------\n")
                    print("Applying doubling map")

                # Convert to a type A_{2n-1}^{(2)} RC
                dims = self.cur_dims[:]
                dims.insert(0, [r, cur_crystal.parent().s()])
                KRT = TensorProductOfKirillovReshetikhinTableaux(['A', 2*self.n-1, 2], dims)
                # Convert the n-th partition into a regular rigged partition
                self.ret_rig_con[-1] = RiggedPartition(self.ret_rig_con[-1]._list,
                                                       self.ret_rig_con[-1].rigging,
                                                       self.ret_rig_con[-1].vacancy_numbers)
                # Placeholder element
                elt = KRT(*[C.module_generators[0] for C in KRT.crystals])
                bij = KRTToRCBijectionTypeA2Odd(elt)
                bij.ret_rig_con = KRT.rigged_configurations()(*self.ret_rig_con, use_vacancy_numbers=True)
                bij.cur_path = self.cur_path
                bij.cur_dims = self.cur_dims
                for i in range(len(self.cur_dims)):
                    if bij.cur_dims[i][0] != self.n:
                        bij.cur_dims[i][1] *= 2
                for i in range(self.n-1):
                    for j in range(len(bij.ret_rig_con[i])):
                        bij.ret_rig_con[i]._list[j] *= 2
                        bij.ret_rig_con[i].rigging[j] *= 2
                        bij.ret_rig_con[i].vacancy_numbers[j] *= 2

                # Perform the type A_{2n-1}^{(2)} bijection
                r = cur_crystal.parent().r()
                # Iterate through the columns
                for col_number, cur_column in enumerate(reversed(cur_crystal.to_array(False))):
                    bij.cur_path.insert(0, [])  # Prepend an empty list
                    bij.cur_dims.insert(0, [0, 1])

                    # Note that we do not need to worry about iterating over columns
                    #   (see previous note about the data structure).
                    for letter in reversed(cur_column):
                        bij.cur_dims[0][0] += 1
                        val = letter.value  # Convert from a CrystalOfLetter to an Integer

                        if verbose:
                            print("====================")
                            print(repr(TensorProductOfKirillovReshetikhinTableauxElement(self.tp_krt.parent(), bij.cur_path)))
                            print("--------------------")
                            print(repr(bij.ret_rig_con))
                            print("--------------------\n")

                        # Build the next state
                        bij.cur_path[0].insert(0, [letter])  # Prepend the value
                        bij.next_state(val)

                    # If we've split off a column, we need to merge the current column
                    #   to the current crystal tableau
                    if col_number > 0:
                        for i, letter_singleton in enumerate(self.cur_path[0]):
                            bij.cur_path[1][i].insert(0, letter_singleton[0])
                        bij.cur_dims[1][1] += 1
                        bij.cur_path.pop(0)
                        bij.cur_dims.pop(0)

                        # And perform the inverse column splitting map on the RC
                        for a in range(self.n):
                            bij._update_vacancy_nums(a)

                if verbose:
                    print("====================")
                    print(repr(TensorProductOfKirillovReshetikhinTableauxElement(self.tp_krt.parent(), bij.cur_path)))
                    print("--------------------")
                    print(repr(bij.ret_rig_con))
                    print("--------------------\n")
                    print("Applying halving map")

                # Convert back to a type B_n^{(1)}
                for i in range(len(self.cur_dims)):
                    if bij.cur_dims[i][0] != self.n:
                        bij.cur_dims[i][1] //= 2
                for i in range(self.n-1):
                    for j in range(len(bij.ret_rig_con[i])):
                        bij.ret_rig_con[i]._list[j] //= 2
                        bij.ret_rig_con[i].rigging[j] //= 2
                        bij.ret_rig_con[i].vacancy_numbers[j] //= 2
                self.ret_rig_con = self.tp_krt.parent().rigged_configurations()(*bij.ret_rig_con, use_vacancy_numbers=True)
                # Make it mutable so we don't have to keep making copies, at the
                #   end of the bijection, we will make it immutable again
                self.ret_rig_con._set_mutable()
            else:
                # Perform the regular type B_n^{(1)} bijection
                # Iterate through the columns
                for col_number, cur_column in enumerate(reversed(cur_crystal.to_array(False))):
                    self.cur_path.insert(0, [])  # Prepend an empty list
                    self.cur_dims.insert(0, [0, 1])

                    # Note that we do not need to worry about iterating over columns
                    #   (see previous note about the data structure).
                    for letter in reversed(cur_column):
                        self.cur_dims[0][0] += 1
                        val = letter.value  # Convert from a CrystalOfLetter to an Integer

                        if verbose:
                            print("====================")
                            print(repr(TensorProductOfKirillovReshetikhinTableauxElement(self.tp_krt.parent(), self.cur_path)))
                            print("--------------------")
                            print(repr(self.ret_rig_con))
                            print("--------------------\n")

                        # Build the next state
                        self.cur_path[0].insert(0, [letter])  # Prepend the value
                        self.next_state(val)

                    # If we've split off a column, we need to merge the current column
                    #   to the current crystal tableau
                    if col_number > 0:
                        if verbose:
                            print("====================")
                            print(repr(TensorProductOfKirillovReshetikhinTableauxElement(self.tp_krt.parent(), self.cur_path)))
                            print("--------------------")
                            print(repr(self.ret_rig_con))
                            print("--------------------\n")
                            print("Applying column merge")

                        for i, letter_singleton in enumerate(self.cur_path[0]):
                            self.cur_path[1][i].insert(0, letter_singleton[0])
                        self.cur_dims[1][1] += 1
                        self.cur_path.pop(0)
                        self.cur_dims.pop(0)

                        # And perform the inverse column splitting map on the RC
                        for a in range(self.n):
                            self._update_vacancy_nums(a)
        self.ret_rig_con.set_immutable()  # Return it to immutable
        return self.ret_rig_con

    def next_state(self, val):
        r"""
        Build the next state for type `B_n^{(1)}`.

        TESTS::

            sage: KRT = crystals.TensorProductOfKirillovReshetikhinTableaux(['B', 3, 1], [[2,1]])
            sage: from sage.combinat.rigged_configurations.bij_type_B import KRTToRCBijectionTypeB
            sage: bijection = KRTToRCBijectionTypeB(KRT(pathlist=[[-1,2]]))
            sage: bijection.cur_path.insert(0, [])
            sage: bijection.cur_dims.insert(0, [0, 1])
            sage: bijection.cur_path[0].insert(0, [3])
            sage: bijection.next_state(3)
        """
        n = self.n
        tableau_height = len(self.cur_path[0]) - 1

        # If it is a regular value, we follow the A_n rules
        if val > 0:
            KRTToRCBijectionTypeA.next_state(self, val)
            return

        pos_val = -val

        # Special case for 0
        if pos_val == 0:
            if len(self.ret_rig_con[pos_val - 1]) > 0:
                max_width = self.ret_rig_con[n-1][0]
            else:
                max_width = 1
            max_width = self.ret_rig_con[n-1].insert_cell(max_width)
            width_n = max_width + 1
            max_width = max_width // 2

            # Check to see if we need to make the new string quasi-singular
            if tableau_height != n-1:
                max_width = self.ret_rig_con[n-2].insert_cell(max_width)
            else:
                max_width = -1
            self._update_vacancy_nums(n - 1)
            self._update_partition_values(n - 1)

            # Check if we need to make the new string at n quasi-singular
            p = self.ret_rig_con[n-1]
            num_rows = len(p)
            # Note that max width is 1 less than the corresponding string length
            if max_width*2 + 1 != width_n:
                for i in range(num_rows):
                    if p._list[i] == width_n:
                        j = i+1
                        while j < num_rows and p._list[j] == width_n \
                          and p.vacancy_numbers[j] == p.rigging[j]:
                            j += 1
                        p.rigging[j-1] -= 1
                        break

            # Follow regular A_n rules
            for a in reversed(range(tableau_height, n-2)):
                max_width = self.ret_rig_con[a].insert_cell(max_width)
                self._update_vacancy_nums(a + 1)
                self._update_partition_values(a + 1)
            self._update_vacancy_nums(tableau_height)
            self._update_partition_values(tableau_height)
            if tableau_height > 0:
                self._update_vacancy_nums(tableau_height-1)
                self._update_partition_values(tableau_height-1)
            return

        # Always add a cell to the first singular value in the first
        #   tableau we are updating.
        if len(self.ret_rig_con[pos_val - 1]) > 0:
            max_width = self.ret_rig_con[pos_val - 1][0] + 1
        else:
            max_width = 0

        # Add cells similar to type A_n but we move to the right until n
        for a in range(pos_val - 1, n - 1):
            max_width = self.ret_rig_con[a].insert_cell(max_width)

        # Handle the special behavior at n
        if pos_val != n:
            max_width = max_width * 2

        # Find either the quasi-singular string and the next largest singular,
        #   the largest singular string if smaller than the max width
        #   or the two largest singular strings
        singular_max_width = False
        case_QS = False
        # Note, case_QS and singular_max_width will never both be True
        p = self.ret_rig_con[n-1]
        num_rows = len(p)
        width_n = 0
        for i in range(num_rows + 1):
            if i == num_rows:
                if case_QS:
                    # If we are in case (QS), we will be adding a box
                    p._list.append(1)
                    p.vacancy_numbers.append(None)
                    p.rigging.append(None)
                    width_n = 1
                    max_width = 0
                elif not singular_max_width:
                    # If we have not found a (quasi)singular string, we must add 2 boxes
                    # Go through our partition until we find a length of greater than 2
                    j = len(p._list) - 1
                    while j >= 0 and p._list[j] <= 2:
                        j -= 1
                    p._list.insert(j+1, 2)
                    p.vacancy_numbers.insert(j+1, None)
                    p.rigging.insert(j+1, None)
                    max_width = 0
                break
            elif p.vacancy_numbers[i] == p.rigging[i]:
                if p._list[i] < max_width:
                    if singular_max_width:
                        width_n = p._list[i]
                        break

                    max_width = p._list[i]
                    if case_QS:
                        p._list[i] += 1
                        p.rigging[i] = None
                        width_n = max_width + 1
                    else:
                        # Add 2 boxes
                        j = i - 1
                        while j >= 0 and p._list[j] <= max_width + 2:
                            p.rigging[j+1] = p.rigging[j]  # Shuffle it along
                            j -= 1
                        p._list.pop(i)
                        p._list.insert(j+1, max_width + 2)
                        p.rigging[j+1] = None
                    break

                if p._list[i] == max_width and not singular_max_width:
                    p._list[i] += 1  # We always at least add a box to the first singular value
                    p.rigging[i] = None
                    if case_QS:
                        width_n = p._list[i]
                        break
                    singular_max_width = True
                elif p._list[i] == max_width + 1 and not case_QS:
                    # If we can't add 2 boxes, we must be in case (QS)
                    p._list[i] += 1
                    p.rigging[i] = None
                    width_n = max_width
                    case_QS = True
            elif p.vacancy_numbers[i] - 1 == p.rigging[i] and not case_QS and not singular_max_width and p._list[i] <= max_width:
                case_QS = True
                max_width = p._list[i]
                p._list[i] += 1
                p.rigging[i] = None

        if singular_max_width:
            # There are 2 possibilities, case (S) and case (QS), we might need
            #   to attempt both
            # Make a *deep* copy of the element
            cp = self.ret_rig_con.__copy__()
            for i, rp in enumerate(cp):
                cp[i] = rp._clone()
            # We attempt case (S) first
            self._insert_cell_case_S(p)

        max_width = max_width // 2

        # We need to do the next partition in order to determine the step at n
        if tableau_height != n-1:
            max_width = self.ret_rig_con[n-2].insert_cell(max_width)
        else:
            max_width = -1

        self._update_vacancy_nums(n - 1)
        self._update_partition_values(n - 1)

        # If we need to make the smaller added string quasisingular
        # Note that max width is 1 less than the corresponding string length
        if case_QS and max_width*2 + 1 != width_n:
            for i in range(num_rows):
                if p._list[i] == width_n:
                    j = i+1
                    while j < num_rows and p._list[j] == width_n \
                      and p.vacancy_numbers[j] == p.rigging[j]:
                        j += 1
                    p.rigging[j-1] -= 1
                    break

        # Continue back following the regular A_n rules
        for a in reversed(range(tableau_height, n - 2)):
            max_width = self.ret_rig_con[a].insert_cell(max_width)
            self._update_vacancy_nums(a + 1)
            self._update_partition_values(a + 1)

        # Update the final rigged partitions
        if tableau_height < n:
            self._update_vacancy_nums(tableau_height)
            self._update_partition_values(tableau_height)

        assert pos_val > 0
        if pos_val <= tableau_height:
            for a in range(pos_val-1, tableau_height):
                self._update_vacancy_nums(a)
                self._update_partition_values(a)
            if pos_val > 1:
                self._update_vacancy_nums(pos_val - 2)
                self._update_partition_values(pos_val - 2)
        elif tableau_height > 0:
            self._update_vacancy_nums(tableau_height - 1)
            self._update_partition_values(tableau_height - 1)

        if singular_max_width:
            try:
                self.ret_rig_con.check()
            except Exception:
                self.other_outcome(cp, pos_val, width_n)

    def other_outcome(self, rc, pos_val, width_n):
        r"""
        Do the other case `(QS)` possibility.

        This arises from the ambiguity when we found a singular string at the
        max width in `\nu^{(n)}`. We had first attempted case `(S)`, and if
        that resulted in an invalid rigged configuration, we now
        finish the bijection using case `(QS)`.

        EXAMPLES::

            sage: RC = RiggedConfigurations(['B',3,1], [[2,1],[1,2]])
            sage: rc = RC(partition_list=[[2,1], [2,1,1], [5,1]])
            sage: t = rc.to_tensor_product_of_kirillov_reshetikhin_tableaux()
            sage: t.to_rigged_configuration() == rc # indirect doctest
            True
        """
        n = self.n
        tableau_height = len(self.cur_path[0]) - 1
        self.ret_rig_con = rc

        # We need to do the next partition in order to determine the step at n
        max_width = self.ret_rig_con[n-2].insert_cell(width_n // 2)

        # We now attempt case (QS)
        case_QS = False
        p = self.ret_rig_con[n-1]
        num_rows = len(p)
        for i in range(len(p._list)):
            if p._list[i] == width_n:
                p._list[i] += 1
                p.rigging[i] = None
                case_QS = True
                break
        if not case_QS:  # we have not added a box yet
            p._list.append(1)
            p.rigging.append(None)
            p.vacancy_numbers.append(None)
            case_QS = True
        width_n += 1

        self._update_vacancy_nums(n - 1)
        self._update_partition_values(n - 1)

        # If we need to make the smaller added string quasisingular
        # Note that max width is 1 less than the corresponding string length
        if case_QS and max_width*2 + 1 != width_n:
            for i in range(num_rows):
                if p._list[i] == width_n:
                    j = i+1
                    while j < num_rows and p._list[j] == width_n \
                      and p.vacancy_numbers[j] == p.rigging[j]:
                        j += 1
                    p.rigging[j-1] -= 1
                    break

        # Continue back following the regular A_n rules
        for a in reversed(range(tableau_height, n - 2)):
            max_width = self.ret_rig_con[a].insert_cell(max_width)
            self._update_vacancy_nums(a + 1)
            self._update_partition_values(a + 1)

        # Update the final rigged partitions
        if tableau_height < n:
            self._update_vacancy_nums(tableau_height)
            self._update_partition_values(tableau_height)

        assert pos_val > 0
        if pos_val <= tableau_height:
            for a in range(pos_val-1, tableau_height):
                self._update_vacancy_nums(a)
                self._update_partition_values(a)
            if pos_val > 1:
                self._update_vacancy_nums(pos_val - 2)
                self._update_partition_values(pos_val - 2)
        elif tableau_height > 0:
            self._update_vacancy_nums(tableau_height - 1)
            self._update_partition_values(tableau_height - 1)


class RCToKRTBijectionTypeB(RCToKRTBijectionTypeC):
    r"""
    Specific implementation of the bijection from rigged configurations to
    tensor products of KR tableaux for type `B_n^{(1)}`.
    """

    def run(self, verbose=False, build_graph=False):
        """
        Run the bijection from rigged configurations to tensor product of KR
        tableaux for type `B_n^{(1)}`.

        INPUT:

        - ``verbose`` -- boolean (default: ``False``); display each step in the
          bijection
        - ``build_graph`` -- boolean (default: ``False``); build the graph of each
          step of the bijection

        EXAMPLES::

            sage: RC = RiggedConfigurations(['B', 3, 1], [[2, 1]])
            sage: from sage.combinat.rigged_configurations.bij_type_B import RCToKRTBijectionTypeB
            sage: RCToKRTBijectionTypeB(RC(partition_list=[[1],[1,1],[1]])).run()
            [[3], [0]]

            sage: RC = RiggedConfigurations(['B', 3, 1], [[3, 1]])
            sage: x = RC(partition_list=[[],[1],[1]])
            sage: RCToKRTBijectionTypeB(x).run()
            [[1], [3], [-2]]
            sage: bij = RCToKRTBijectionTypeB(x)
            sage: bij.run(build_graph=True)
            [[1], [3], [-2]]
            sage: bij._graph
            Digraph on 6 vertices
        """
        from sage.combinat.crystals.letters import CrystalOfLetters
        letters = CrystalOfLetters(self.rigged_con.parent()._cartan_type.classical())

        # This is technically bad, but because the first thing we do is append
        #   an empty list to ret_crystal_path, we correct this. We do it this
        #   way so that we do not have to remove an empty list after the
        #   bijection has been performed.
        ret_crystal_path = []

        for dim in self.rigged_con.parent().dims:
            ret_crystal_path.append([])

            # Check to see if we are a spinor
            if dim[0] == self.n:
                # Perform the spinor bijection by converting to type A_{2n-1}^{(2)}
                #   doing the bijection there and pulling back

                from sage.combinat.rigged_configurations.bij_type_A2_odd import RCToKRTBijectionTypeA2Odd
                from sage.combinat.rigged_configurations.rigged_configurations import RiggedConfigurations
                from sage.combinat.rigged_configurations.rigged_partition import RiggedPartition, RiggedPartitionTypeB

                # Convert to a type A_{2n-1}^{(2)} RC
                RC = RiggedConfigurations(['A', 2*self.n-1, 2], self.cur_dims)
                if verbose:
                    print("====================")
                    print(repr(RC(*self.cur_partitions, use_vacancy_numbers=True)))
                    print("--------------------")
                    print(ret_crystal_path)
                    print("--------------------\n")
                    print("Applying doubling map\n")
                # Convert the n-th partition into a regular rigged partition
                self.cur_partitions[-1] = RiggedPartition(self.cur_partitions[-1]._list,
                                                          self.cur_partitions[-1].rigging,
                                                          self.cur_partitions[-1].vacancy_numbers)

                bij = RCToKRTBijectionTypeA2Odd(RC(*self.cur_partitions, use_vacancy_numbers=True))
                for i in range(len(self.cur_dims)):
                    if bij.cur_dims[i][0] != self.n:
                        bij.cur_dims[i][1] *= 2
                for i in range(self.n-1):
                    for j in range(len(bij.cur_partitions[i])):
                        bij.cur_partitions[i]._list[j] *= 2
                        bij.cur_partitions[i].rigging[j] *= 2
                        bij.cur_partitions[i].vacancy_numbers[j] *= 2

                if build_graph:
                    y = self.rigged_con.parent()(*[x._clone() for x in self.cur_partitions], use_vacancy_numbers=True)
                    self._graph.append([self._graph[-1][1], (y, len(self._graph)), '2x'])

                # Perform the type A_{2n-1}^{(2)} bijection

                # Iterate over each column
                for dummy_var in range(dim[1]):
                    # Split off a new column if necessary
                    if bij.cur_dims[0][1] > 1:
                        bij.cur_dims[0][1] -= 1
                        bij.cur_dims.insert(0, [dim[0], 1])

                        # Perform the corresponding splitting map on rigged configurations
                        # All it does is update the vacancy numbers on the RC side
                        for a in range(self.n):
                            bij._update_vacancy_numbers(a)

                        if build_graph:
                            y = self.rigged_con.parent()(*[x._clone() for x in self.cur_partitions], use_vacancy_numbers=True)
                            self._graph.append([self._graph[-1][1], (y, len(self._graph)), 'ls'])

                    while bij.cur_dims[0][0]:  # > 0:
                        if verbose:
                            print("====================")
                            print(repr(RC(*bij.cur_partitions, use_vacancy_numbers=True)))
                            print("--------------------")
                            print(ret_crystal_path)
                            print("--------------------\n")

                        ht = bij.cur_dims[0][0]
                        bij.cur_dims[0][0] = bij._next_index(ht)
                        b = bij.next_state(ht)
                        # Make sure we have a crystal letter
                        ret_crystal_path[-1].append(letters(b))  # Append the rank

                        if build_graph:
                            y = self.rigged_con.parent()(*[x._clone() for x in self.cur_partitions], use_vacancy_numbers=True)
                            self._graph.append([self._graph[-1][1], (y, len(self._graph)), letters(b)])

                    bij.cur_dims.pop(0)  # Pop off the leading column

                self.cur_dims.pop(0)  # Pop off the spin rectangle

                self.cur_partitions = bij.cur_partitions
                # Convert the n-th partition back into the special type B one
                self.cur_partitions[-1] = RiggedPartitionTypeB(self.cur_partitions[-1])

                # Convert back to a type B_n^{(1)}
                if verbose:
                    print("====================")
                    print(repr(self.rigged_con.parent()(*bij.cur_partitions, use_vacancy_numbers=True)))
                    print("--------------------")
                    print(ret_crystal_path)
                    print("--------------------\n")
                    print("Applying halving map\n")

                for i in range(self.n-1):
                    for j in range(len(self.cur_partitions[i])):
                        self.cur_partitions[i]._list[j] //= 2
                        self.cur_partitions[i].rigging[j] //= 2
                        self.cur_partitions[i].vacancy_numbers[j] //= 2

                if build_graph:
                    y = self.rigged_con.parent()(*[x._clone() for x in self.cur_partitions], use_vacancy_numbers=True)
                    self._graph.append([self._graph[-1][1], (y, len(self._graph)), '1/2x'])
            else:
                # Perform the regular type B_n^{(1)} bijection

                # Iterate over each column
                for dummy_var in range(dim[1]):
                    # Split off a new column if necessary
                    if self.cur_dims[0][1] > 1:
                        if verbose:
                            print("====================")
                            print(repr(self.rigged_con.parent()(*self.cur_partitions, use_vacancy_numbers=True)))
                            print("--------------------")
                            print(ret_crystal_path)
                            print("--------------------\n")
                            print("Applying column split")

                        self.cur_dims[0][1] -= 1
                        self.cur_dims.insert(0, [dim[0], 1])

                        # Perform the corresponding splitting map on rigged configurations
                        # All it does is update the vacancy numbers on the RC side
                        for a in range(self.n):
                            self._update_vacancy_numbers(a)

                        if build_graph:
                            y = self.rigged_con.parent()(*[x._clone() for x in self.cur_partitions], use_vacancy_numbers=True)
                            self._graph.append([self._graph[-1][1], (y, len(self._graph)), '2x'])

                    while self.cur_dims[0][0]:  # > 0:
                        if verbose:
                            print("====================")
                            print(repr(self.rigged_con.parent()(*self.cur_partitions, use_vacancy_numbers=True)))
                            print("--------------------")
                            print(ret_crystal_path)
                            print("--------------------\n")

                        self.cur_dims[0][0] -= 1  # This takes care of the indexing
                        b = self.next_state(self.cur_dims[0][0])

                        # Make sure we have a crystal letter
                        ret_crystal_path[-1].append(letters(b))  # Append the rank

                        if build_graph:
                            y = self.rigged_con.parent()(*[x._clone() for x in self.cur_partitions], use_vacancy_numbers=True)
                            self._graph.append([self._graph[-1][1], (y, len(self._graph)), letters(b)])

                    self.cur_dims.pop(0)  # Pop off the leading column

        if build_graph:
            self._graph.pop(0)  # Remove the dummy at the start
            from sage.graphs.digraph import DiGraph
            from sage.graphs.dot2tex_utils import have_dot2tex
            self._graph = DiGraph(self._graph)
            if have_dot2tex():
                self._graph.set_latex_options(format='dot2tex', edge_labels=True)

        return self.KRT(pathlist=ret_crystal_path)

    def next_state(self, height):
        r"""
        Build the next state for type `B_n^{(1)}`.

        TESTS::

            sage: RC = RiggedConfigurations(['B', 3, 1], [[2, 1]])
            sage: from sage.combinat.rigged_configurations.bij_type_B import RCToKRTBijectionTypeB
            sage: bijection = RCToKRTBijectionTypeB(RC(partition_list=[[1],[1,1],[1]]))
            sage: bijection.next_state(0)
            0
        """
        n = self.n
        ell = [None] * (2*n)
        case_S = False
        case_Q = False
        b = None

        # Calculate the rank and ell values

        last_size = 0
        for a in range(height, n-1):
            ell[a] = self._find_singular_string(self.cur_partitions[a], last_size)

            if ell[a] is None:
                b = a + 1
                break
            else:
                last_size = self.cur_partitions[a][ell[a]]

        # Special case for n
        if b is None:
            last_size = 2 * last_size - 1
            partition = self.cur_partitions[n-1]
            # Modified version of _find_singular_string()
            for i in reversed(range(len(partition))):
                if partition[i] == last_size \
                  and partition.vacancy_numbers[i] == partition.rigging[i]:
                    case_Q = True
                    ell[n-1] = i
                elif partition[i] > last_size:
                    if not case_Q and partition.vacancy_numbers[i] - 1 == partition.rigging[i]:
                        case_Q = True
                        # Check if the block is singular as well
                        block_size = partition[i]
                        for j in reversed(range(i)):
                            if partition[j] != block_size:
                                break
                            elif partition.vacancy_numbers[j] == partition.rigging[j]:
                                case_Q = False
                                ell[2*n-1] = j
                                last_size = partition[j]
                                case_S = True
                                break
                        if not case_Q:  # We found a singular string above the quasi-singular one
                            break
                        ell[n-1] = i
                        last_size = partition[i]
                        # Now check for case QS
                    elif partition.vacancy_numbers[i] == partition.rigging[i]:
                        ell[2*n-1] = i
                        last_size = partition[i]
                        case_S = True
                        break

            if ell[2*n-1] is None:
                if not case_Q:
                    b = n
                else:
                    b = 0

        if b is None:
            # Now go back
            last_size = (last_size + 1) // 2
            for a in reversed(range(n - 1)):
                # Modified form of _find_singular_string
                end = ell[a]
                if a < height:
                    end = len(self.cur_partitions[a])
                for i in reversed(range(end)):
                    if self.cur_partitions[a][i] >= last_size and \
                      self.cur_partitions[a].vacancy_numbers[i] == self.cur_partitions[a].rigging[i]:
                        ell[n + a] = i
                        break

                if ell[n + a] is None:
                    b = -(a + 2)
                    break
                else:
                    last_size = self.cur_partitions[a][ell[n + a]]

        if b is None:
            b = -1

        # Determine the new rigged configuration by removing boxes from the
        #   selected string and then making the new string singular

        # Determine if we need to make the n-th string quasisingular
        make_quasisingular = case_Q and case_S and \
                (ell[2*n-2] is None
                 or self.cur_partitions[n-1][ell[2*n-1]]
                    < 2*self.cur_partitions[n-2][ell[2*n-2]])

        row_num = self.cur_partitions[0].remove_cell(ell[0])
        row_num_bar = self.cur_partitions[0].remove_cell(ell[n])
        for a in range(1, n-1):
            row_num_next = self.cur_partitions[a].remove_cell(ell[a])
            row_num_bar_next = self.cur_partitions[a].remove_cell(ell[n+a])

            self._update_vacancy_numbers(a - 1)
            if row_num is not None:
                self.cur_partitions[a-1].rigging[row_num] = self.cur_partitions[a-1].vacancy_numbers[row_num]
            if row_num_bar is not None:
                self.cur_partitions[a-1].rigging[row_num_bar] = self.cur_partitions[a-1].vacancy_numbers[row_num_bar]
            row_num = row_num_next
            row_num_bar = row_num_bar_next

        if case_Q:
            if case_S:
                row_num_next = self.cur_partitions[n-1].remove_cell(ell[n-1])
                row_num_bar_next = self.cur_partitions[n-1].remove_cell(ell[2*n-1])
            else:
                row_num_next = self.cur_partitions[n-1].remove_cell(ell[n-1])
                row_num_bar_next = None
        elif case_S:
            row_num_next = self.cur_partitions[n-1].remove_cell(ell[2*n-1], 2)
            row_num_bar_next = None
        else:
            row_num_next = None
            row_num_bar_next = None

        self._update_vacancy_numbers(n - 2)
        if row_num is not None:
            self.cur_partitions[n-2].rigging[row_num] = self.cur_partitions[n-2].vacancy_numbers[row_num]
        if row_num_bar is not None:
            self.cur_partitions[n-2].rigging[row_num_bar] = self.cur_partitions[n-2].vacancy_numbers[row_num_bar]

        self._update_vacancy_numbers(n - 1)
        if row_num_next is not None:
            self.cur_partitions[n-1].rigging[row_num_next] = self.cur_partitions[n-1].vacancy_numbers[row_num_next]
        if row_num_bar_next is not None:  # If we enter here, it means case (Q, S) holds
            vac_num = self.cur_partitions[n-1].vacancy_numbers[row_num_bar_next]
            self.cur_partitions[n-1].rigging[row_num_bar_next] = vac_num
            if make_quasisingular:
                block_len = self.cur_partitions[n-1][row_num_bar_next]
                j = row_num_bar_next + 1
                length = len(self.cur_partitions[n-1])
                # Find the place for the quasisingular rigging
                while j < length and self.cur_partitions[n-1][j] == block_len \
                  and self.cur_partitions[n-1].rigging[j] == vac_num:
                    j += 1
                self.cur_partitions[n-1].rigging[j-1] = vac_num - 1

        return b



# sage_setup: distribution = sagemath-combinat
# sage.doctest: needs sage.combinat sage.graphs sage.modules
"""
Crystals of Kac modules of the general-linear Lie superalgebra
"""

#*****************************************************************************
#       Copyright (C) 2017 Travis Scrimshaw <tcscrims at gmail.com>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 2 of the License, or
# (at your option) any later version.
#                  http://www.gnu.org/licenses/
#*****************************************************************************

from sage.structure.parent import Parent
from sage.structure.element_wrapper import ElementWrapper
from sage.structure.unique_representation import UniqueRepresentation
from sage.rings.integer_ring import ZZ

from sage.categories.regular_supercrystals import RegularSuperCrystals
from sage.combinat.crystals.tensor_product import CrystalOfTableaux
from sage.combinat.root_system.cartan_type import CartanType
from sage.combinat.partition import _Partitions


class CrystalOfOddNegativeRoots(UniqueRepresentation, Parent):
    r"""
    Crystal of the set of odd negative roots.

    Let `\mathfrak{g}` be the general-linear Lie superalgebra
    `\mathfrak{gl}(m|n)`. This is the crystal structure on the set of
    negative roots as given by [Kwon2012]_.

    More specifically, this is the crystal basis of the subalgebra
    of `U_q^-(\mathfrak{g})` generated by `f_{\alpha}`, where `\alpha`
    ranges over all odd positive roots. As `\QQ(q)`-modules, we have

    .. MATH::

        U_q^-(\mathfrak{g}) \cong
        K \otimes U^-_q(\mathfrak{gl}_m \oplus \mathfrak{gl}_n).

    EXAMPLES::

        sage: S = crystals.OddNegativeRoots(['A', [2,1]])
        sage: mg = S.module_generator(); mg
        {}
        sage: mg.f(0)
        {-e[-1]+e[1]}
        sage: mg.f_string([0,-1,0,1,2,1,0])
        {-e[-2]+e[3], -e[-1]+e[1], -e[-1]+e[2]}
    """
    @staticmethod
    def __classcall_private__(cls, cartan_type):
        """
        Normalize input to ensure a unique representation.

        TESTS::

            sage: S1 = crystals.OddNegativeRoots(['A', [2,1]])
            sage: S2 = crystals.OddNegativeRoots(CartanType(['A', [2,1]]))
            sage: S1 is S2
            True
        """
        return super().__classcall__(cls, CartanType(cartan_type))

    def __init__(self, cartan_type):
        """
        Initialize ``self``.

        TESTS::

            sage: S = crystals.OddNegativeRoots(['A', [2,1]])
            sage: TestSuite(S).run()
        """
        self._cartan_type = cartan_type
        Parent.__init__(self, category=RegularSuperCrystals())
        self.module_generators = (self.element_class(self, frozenset()),)

    def _repr_(self):
        """
        Return a string representation of ``self``.

        EXAMPLES::

            sage: crystals.OddNegativeRoots(['A', [2,1]])
            Crystal of odd negative roots of type ['A', [2, 1]]
        """
        return "Crystal of odd negative roots of type {}".format(self._cartan_type)

    def module_generator(self):
        """
        Return the module generator of ``self``.

        EXAMPLES::

            sage: S = crystals.OddNegativeRoots(['A', [2,1]])
            sage: S.module_generator()
            {}
        """
        return self.module_generators[0]

    class Element(ElementWrapper):
        """
        An element of the crystal of odd negative roots.

        TESTS:

        Check that `e_i` and `f_i` are psuedo-inverses::

            sage: S = crystals.OddNegativeRoots(['A', [2,1]])
            sage: for x in S:
            ....:     for i in S.index_set():
            ....:         y = x.f(i)
            ....:         assert y is None or y.e(i) == x

        Check that we obtain the entire powerset of negative odd roots::

            sage: S = crystals.OddNegativeRoots(['A', [2,3]])
            sage: S.cardinality()
            4096
            sage: 2^len(S.weight_lattice_realization().positive_odd_roots())
            4096
        """

        def _repr_(self):
            r"""
            Return a string representation of ``self``.

            EXAMPLES::

                sage: S = crystals.OddNegativeRoots(['A', [2,2]])
                sage: mg = S.module_generator(); mg
                {}
                sage: mg.f(0)
                {-e[-1]+e[1]}
                sage: mg.f_string([0,-1,0])
                {-e[-2]+e[1], -e[-1]+e[1]}
            """
            return ('{'
                    + ", ".join("-e[{}]+e[{}]".format(*i)
                                for i in sorted(self.value))
                    + '}')

        def _latex_(self):
            r"""
            Return a latex representation of ``self``.

            EXAMPLES::

                sage: S = crystals.OddNegativeRoots(['A', [2,2]])
                sage: mg = S.module_generator()
                sage: latex(mg)
                \{\}
                sage: latex(mg.f(0))
                \{-e_{-1}+e_{1}\}
                sage: latex(mg.f_string([0,-1,0]))
                \{-e_{-2}+e_{1}, -e_{-1}+e_{1}\}
            """
            return (r'\{'
                    + ", ".join("-e_{{{}}}+e_{{{}}}".format(*i)
                                for i in sorted(self.value))
                    + r'\}')

        def e(self, i):
            r"""
            Return the action of the crystal operator `e_i` on ``self``.

            EXAMPLES::

                sage: S = crystals.OddNegativeRoots(['A', [2,2]])
                sage: mg = S.module_generator()
                sage: mg.e(0)
                sage: mg.e(1)
                sage: b = mg.f_string([0,1,2,-1,0])
                sage: b.e(-1)
                sage: b.e(0)
                {-e[-2]+e[3]}
                sage: b.e(1)
                sage: b.e(2)
                {-e[-2]+e[2], -e[-1]+e[1]}
                sage: b.e_string([2,1,0,-1,0])
                {}
            """
            if i == 0:
                if (-1,1) not in self.value:
                    return None
                return type(self)(self.parent(), self.value.difference([(-1,1)]))

            count = 0
            act_val = None
            if i < 0:
                lst = sorted(self.value, key=lambda x: (x[1], -x[0]))
                for val in lst:
                    # We don't have to check val[1] because this is an odd root
                    if val[0] == i - 1:
                        if count == 0:
                            act_val = val
                        else:
                            count -= 1
                    elif val[0] == i:
                        count += 1
                if act_val is None:
                    return None
                ret = self.value.difference([act_val]).union([(i, act_val[1])])
                return type(self)(self.parent(), ret)

            # else i > 0
            lst = sorted(self.value, key=lambda x: (-x[0], -x[1]))
            for val in reversed(lst):
                # We don't have to check val[0] because this is an odd root
                if val[1] == i + 1:
                    if count == 0:
                        act_val = val
                    else:
                        count -= 1
                elif val[1] == i:
                    count += 1
            if act_val is None:
                return None
            ret = self.value.difference([act_val]).union([(act_val[0], i)])
            return type(self)(self.parent(), ret)

        def f(self, i):
            r"""
            Return the action of the crystal operator `f_i` on ``self``.

            EXAMPLES::

                sage: S = crystals.OddNegativeRoots(['A', [2,2]])
                sage: mg = S.module_generator()
                sage: mg.f(0)
                {-e[-1]+e[1]}
                sage: mg.f(1)
                sage: b = mg.f_string([0,1,2,-1,0]); b
                {-e[-2]+e[3], -e[-1]+e[1]}
                sage: b.f(-2)
                {-e[-3]+e[3], -e[-1]+e[1]}
                sage: b.f(-1)
                sage: b.f(0)
                sage: b.f(1)
                {-e[-2]+e[3], -e[-1]+e[2]}
            """
            if i == 0:
                if (-1,1) in self.value:
                    return None
                return type(self)(self.parent(), self.value.union([(-1,1)]))

            count = 0
            act_val = None
            if i < 0:
                lst = sorted(self.value, key=lambda x: (x[1], -x[0]))
                for val in reversed(lst):
                    # We don't have to check val[1] because this is an odd root
                    if val[0] == i:
                        if count == 0:
                            act_val = val
                        else:
                            count -= 1
                    elif val[0] == i - 1:
                        count += 1
                if act_val is None:
                    return None
                ret = self.value.difference([act_val]).union([(i-1, act_val[1])])
                return type(self)(self.parent(), ret)

            # else i > 0
            lst = sorted(self.value, key=lambda x: (-x[0], -x[1]))
            for val in lst:
                # We don't have to check val[0] because this is an odd root
                if val[1] == i:
                    if count == 0:
                        act_val = val
                    else:
                        count -= 1
                elif val[1] == i + 1:
                    count += 1
            if act_val is None:
                return None
            ret = self.value.difference([act_val]).union([(act_val[0], i+1)])
            return type(self)(self.parent(), ret)

        def epsilon(self, i):
            r"""
            Return `\varepsilon_i` of ``self``.

            EXAMPLES::

                sage: S = crystals.OddNegativeRoots(['A', [2,2]])
                sage: mg = S.module_generator()
                sage: [mg.epsilon(i) for i in S.index_set()]
                [0, 0, 0, 0, 0]
                sage: b = mg.f_string([0,1,0,-1,0,-1,-2,-2]); b
                {-e[-3]+e[1], -e[-3]+e[2], -e[-1]+e[1]}
                sage: [b.epsilon(i) for i in S.index_set()]
                [2, 0, 1, 0, 0]
                sage: b = mg.f_string([0,1,0,-1,0,-1,-2,-2,2,-1,0]); b
                {-e[-3]+e[1], -e[-3]+e[3], -e[-2]+e[1], -e[-1]+e[1]}
                sage: [b.epsilon(i) for i in S.index_set()]
                [1, 0, 1, 0, 1]

            TESTS::

                sage: S = crystals.OddNegativeRoots(['A', [2,1]])
                sage: def count_e(x, i):
                ....:     ret = -1
                ....:     while x is not None:
                ....:         x = x.e(i)
                ....:         ret += 1
                ....:     return ret
                sage: for x in S:
                ....:     for i in S.index_set():
                ....:         assert x.epsilon(i) == count_e(x, i)
            """
            if i == 0:
                return ZZ.one() if (-1,1) in self.value else ZZ.zero()

            count = 0
            ret = 0
            if i < 0:
                lst = sorted(self.value, key=lambda x: (x[1], -x[0]))
                for val in lst:
                    # We don't have to check val[1] because this is an odd root
                    if val[0] == i - 1:
                        if count == 0:
                            ret += 1
                        else:
                            count -= 1
                    elif val[0] == i:
                        count += 1

            else: # i > 0
                lst = sorted(self.value, key=lambda x: (-x[0], -x[1]))
                for val in reversed(lst):
                    # We don't have to check val[0] because this is an odd root
                    if val[1] == i + 1:
                        if count == 0:
                            ret += 1
                        else:
                            count -= 1
                    elif val[1] == i:
                        count += 1
            return ret

        def phi(self, i):
            r"""
            Return `\varphi_i` of ``self``.

            EXAMPLES::

                sage: S = crystals.OddNegativeRoots(['A', [2,2]])
                sage: mg = S.module_generator()
                sage: [mg.phi(i) for i in S.index_set()]
                [0, 0, 1, 0, 0]
                sage: b = mg.f(0)
                sage: [b.phi(i) for i in S.index_set()]
                [0, 1, 0, 1, 0]
                sage: b = mg.f_string([0,1,0,-1,0,-1]); b
                {-e[-2]+e[1], -e[-2]+e[2], -e[-1]+e[1]}
                sage: [b.phi(i) for i in S.index_set()]
                [2, 0, 0, 1, 1]

            TESTS::

                sage: S = crystals.OddNegativeRoots(['A', [2,1]])
                sage: def count_f(x, i):
                ....:     ret = -1
                ....:     while x is not None:
                ....:         x = x.f(i)
                ....:         ret += 1
                ....:     return ret
                sage: for x in S:
                ....:     for i in S.index_set():
                ....:         assert x.phi(i) == count_f(x, i)
            """
            if i == 0:
                return ZZ.zero() if (-1,1) in self.value else ZZ.one()

            count = 0
            ret = 0
            if i < 0:
                lst = sorted(self.value, key=lambda x: (x[1], -x[0]))
                for val in reversed(lst):
                    # We don't have to check val[1] because this is an odd root
                    if val[0] == i:
                        if count == 0:
                            ret += 1
                        else:
                            count -= 1
                    elif val[0] == i - 1:
                        count += 1

            else: # i > 0
                lst = sorted(self.value, key=lambda x: (-x[0], -x[1]))
                for val in lst:
                    # We don't have to check val[0] because this is an odd root
                    if val[1] == i:
                        if count == 0:
                            ret += 1
                        else:
                            count -= 1
                    elif val[1] == i + 1:
                        count += 1
            return ret

        def weight(self):
            r"""
            Return the weight of ``self``.

            EXAMPLES::

                sage: S = crystals.OddNegativeRoots(['A', [2,2]])
                sage: mg = S.module_generator()
                sage: mg.weight()
                (0, 0, 0, 0, 0, 0)
                sage: mg.f_string([0,1,2,-1,-2]).weight()
                (-1, 0, 0, 0, 0, 1)
                sage: mg.f_string([0,1,2,-1,-2,0,1,0,2]).weight()
                (-1, 0, -2, 1, 0, 2)

            TESTS::

                sage: S = crystals.OddNegativeRoots(['A', [2,1]])
                sage: al = S.weight_lattice_realization().simple_roots()
                sage: for x in S:
                ....:     for i in S.index_set():
                ....:         y = x.f(i)
                ....:         assert y is None or x.weight() - al[i] == y.weight()
            """
            WLR = self.parent().weight_lattice_realization()
            e = WLR.basis()
            return WLR.sum(-e[i]+e[j] for (i,j) in self.value)


class CrystalOfKacModule(UniqueRepresentation, Parent):
    r"""
    Crystal of a Kac module.

    Let `\mathfrak{g}` be the general linear Lie superalgebra
    `\mathfrak{gl}(m|n)`. Let `\lambda` and `\mu` be dominant weights
    for `\mathfrak{gl}_m` and `\mathfrak{gl}_n`, respectively.
    Let `K` be the module `K = \langle f_{\alpha} \rangle`,
    where `\alpha` ranges over all odd positive roots. A *Kac module*
    is the `U_q(\mathfrak{g})`-module constructed from the highest
    weight `U_q(\mathfrak{gl}_m \oplus \mathfrak{gl}_n)`-module
    `V(\lambda, \mu)` (induced to a `U_q(\mathfrak{g})`-module in
    the natural way) by

    .. MATH::

        K(\lambda, \mu) := K \otimes_L V(\lambda, \mu),

    where `L` is the subalgebra generated by `e_0` and
    `U_q(\mathfrak{gl}_m \oplus \mathfrak{gl}_n)`.

    The Kac module admits a `U_q(\mathfrak{g})`-crystal structure
    by taking the crystal structure of `K` as given by
    :class:`~sage.combinat.crystals.kac_modules.CrystalOfOddNegativeRoots`
    and the crystal `B(\lambda, \mu)` (the natural crystal structure
    of `V(\lambda, \mu)`).

    .. NOTE::

        Our notation differs slightly from [Kwon2012]_ in that our
        last tableau is transposed.

    EXAMPLES::

        sage: K = crystals.KacModule(['A', [1,2]], [2], [1,1])
        sage: K.cardinality()
        576
        sage: K.cardinality().factor()
        2^6 * 3^2
        sage: len(K.cartan_type().root_system().ambient_space().positive_odd_roots())
        6
        sage: mg = K.module_generator()
        sage: mg
        ({}, [[-2, -2]], [[1], [2]])
        sage: mg.weight()
        (2, 0, 1, 1, 0)
        sage: mg.f(-1)
        ({}, [[-2, -1]], [[1], [2]])
        sage: mg.f(0)
        ({-e[-1]+e[1]}, [[-2, -2]], [[1], [2]])
        sage: mg.f(1)
        sage: mg.f(2)
        ({}, [[-2, -2]], [[1], [3]])

        sage: sorted(K.highest_weight_vectors(), key=str)
        [({-e[-1]+e[3]}, [[-2, -1]], [[1], [2]]),
         ({-e[-1]+e[3]}, [[-2, -2]], [[1], [2]]),
         ({}, [[-2, -2]], [[1], [2]])]

    ::

        sage: K = crystals.KacModule(['A', [1,1]], [2], [1])
        sage: K.cardinality()
        96
        sage: K.cardinality().factor()
        2^5 * 3
        sage: len(K.cartan_type().root_system().ambient_space().positive_odd_roots())
        4

        sage: sorted(K.highest_weight_vectors(), key=str)
        [({-e[-1]+e[2]}, [[-2, -1]], [[1]]),
         ({-e[-1]+e[2]}, [[-2, -2]], [[1]]),
         ({}, [[-2, -2]], [[1]])]
        sage: K.genuine_lowest_weight_vectors()
        (({-e[-2]+e[1], -e[-2]+e[2], -e[-1]+e[1], -e[-1]+e[2]}, [[-1, -1]], [[2]]),)
        sage: sorted(K.lowest_weight_vectors(), key=str)
        [({-e[-1]+e[1], -e[-1]+e[2]}, [[-1, -1]], [[2]]),
         ({-e[-2]+e[1], -e[-2]+e[2], -e[-1]+e[1], -e[-1]+e[2]}, [[-1, -1]], [[2]]),
         ({-e[-2]+e[2], -e[-1]+e[1], -e[-1]+e[2]}, [[-1, -1]], [[1]]),
         ({-e[-2]+e[2], -e[-1]+e[1], -e[-1]+e[2]}, [[-1, -1]], [[2]])]

    REFERENCES:

    - [Kwon2012]_
    """
    @staticmethod
    def __classcall_private__(cls, cartan_type, la, mu):
        """
        Normalize input to ensure a unique representation.

        TESTS::

            sage: K1 = crystals.KacModule(['A', [2,1]], [2,1], [1])
            sage: K2 = crystals.KacModule(CartanType(['A', [2,1]]), (2,1), (1,))
            sage: K1 is K2
            True
        """
        cartan_type = CartanType(cartan_type)
        la = _Partitions(la)
        mu = _Partitions(mu)
        return super().__classcall__(cls, cartan_type, la, mu)

    def __init__(self, cartan_type, la, mu):
        """
        Initialize ``self``.

        TESTS::

            sage: K = crystals.KacModule(['A', [2,1]], [2,1], [1])
            sage: TestSuite(K).run()
        """
        self._cartan_type = cartan_type
        self._la = la
        self._mu = mu
        Parent.__init__(self, category=RegularSuperCrystals())
        self._S = CrystalOfOddNegativeRoots(self._cartan_type)
        self._dual = CrystalOfTableaux(['A', self._cartan_type.m], shape=la)
        self._reg = CrystalOfTableaux(['A', self._cartan_type.n], shape=mu)
        data = (self._S.module_generators[0],
                self._dual.module_generators[0],
                self._reg.module_generators[0])
        self.module_generators = (self.element_class(self, data),)

    def _repr_(self):
        r"""
        Return a string representation of ``self``.

        EXAMPLES::

            sage: crystals.KacModule(['A', [2,1]], [3,1], [1])
            Crystal of Kac module K([3, 1], [1]) of type ['A', [2, 1]]
        """
        return "Crystal of Kac module K({}, {}) of type {}".format(
                    self._la, self._mu, self._cartan_type)

    def module_generator(self):
        """
        Return the module generator of ``self``.

        EXAMPLES::

            sage: K = crystals.KacModule(['A', [2,1]], [2,1], [1])
            sage: K.module_generator()
            ({}, [[-3, -3], [-2]], [[1]])
        """
        return self.module_generators[0]

    class Element(ElementWrapper):
        r"""
        An element of a Kac module crystal.

        TESTS:

        Check that `e_i` and `f_i` are psuedo-inverses::

            sage: K = crystals.KacModule(['A', [2,1]], [2,1], [1])
            sage: for x in K:
            ....:     for i in K.index_set():
            ....:         y = x.f(i)
            ....:         assert y is None or y.e(i) == x
        """

        def _repr_(self):
            """
            Return a string representation of ``self``.

            EXAMPLES::

                sage: K = crystals.KacModule(['A', [2,1]], [2,1], [1])
                sage: mg = K.module_generator(); mg
                ({}, [[-3, -3], [-2]], [[1]])
                sage: mg.f_string([0,1,-2,1,-1,0,-1,-1,1,-2,-2])
                ({-e[-3]+e[2], -e[-1]+e[2]}, [[-2, -1], [-1]], [[2]])
            """
            return repr((self.value[0], to_dual_tableau(self.value[1]), self.value[2]))

        def _latex_(self):
            r"""
            Return a string representation of ``self``.

            EXAMPLES::

                sage: K = crystals.KacModule(['A', [2,1]], [2,1], [1])
                sage: mg = K.module_generator()
                sage: latex(mg)
                \{\}
                 \otimes {\def\lr#1{\multicolumn{1}{|@{\hspace{.6ex}}c@{\hspace{.6ex}}|}{\raisebox{-.3ex}{$#1$}}}
                \raisebox{-.6ex}{$\begin{array}[b]{*{2}c}\cline{1-2}
                \lr{\overline{3}}&\lr{\overline{3}}\\\cline{1-2}
                \lr{\overline{2}}\\\cline{1-1}
                \end{array}$}
                } \otimes {\def\lr#1{\multicolumn{1}{|@{\hspace{.6ex}}c@{\hspace{.6ex}}|}{\raisebox{-.3ex}{$#1$}}}
                \raisebox{-.6ex}{$\begin{array}[b]{*{1}c}\cline{1-1}
                \lr{1}\\\cline{1-1}
                \end{array}$}
                }
                sage: latex(mg.f_string([0,1,-2,1,-1,0,-1,-1,1,-2,-2]))
                \{-e_{-3}+e_{2}, -e_{-1}+e_{2}\}
                 \otimes {\def\lr#1{\multicolumn{1}{|@{\hspace{.6ex}}c@{\hspace{.6ex}}|}{\raisebox{-.3ex}{$#1$}}}
                \raisebox{-.6ex}{$\begin{array}[b]{*{2}c}\cline{1-2}
                \lr{\overline{2}}&\lr{\overline{1}}\\\cline{1-2}
                \lr{\overline{1}}\\\cline{1-1}
                \end{array}$}
                } \otimes {\def\lr#1{\multicolumn{1}{|@{\hspace{.6ex}}c@{\hspace{.6ex}}|}{\raisebox{-.3ex}{$#1$}}}
                \raisebox{-.6ex}{$\begin{array}[b]{*{1}c}\cline{1-1}
                \lr{2}\\\cline{1-1}
                \end{array}$}
                }
            """
            from sage.misc.latex import latex
            return r" \otimes ".join([latex(self.value[0]),
                                      latex_dual(self.value[1]),
                                      latex(self.value[2])])

        def e(self, i):
            r"""
            Return the action of the crystal operator `e_i` on ``self``.

            EXAMPLES::

                sage: K = crystals.KacModule(['A', [2,2]], [2,1], [1])
                sage: mg = K.module_generator()
                sage: mg.e(0)
                sage: mg.e(1)
                sage: mg.e(-1)
                sage: b = mg.f_string([1,0,1,-1,-2,0,1,2,0,-2,-1,-1,-1]); b
                ({-e[-3]+e[2], -e[-2]+e[1], -e[-2]+e[2]}, [[-3, -1], [-2]], [[3]])
                sage: b.e(-2)
                sage: b.e(-1)
                ({-e[-3]+e[2], -e[-2]+e[1], -e[-2]+e[2]}, [[-3, -2], [-2]], [[3]])
                sage: b.e(0)
                sage: b.e(1)
                ({-e[-3]+e[1], -e[-2]+e[1], -e[-2]+e[2]}, [[-3, -1], [-2]], [[3]])
                sage: b.e(2)
                ({-e[-3]+e[2], -e[-2]+e[1], -e[-2]+e[2]}, [[-3, -1], [-2]], [[2]])
            """
            if i == 0:
                x = self.value[0].e(i)
                if x is None:
                    return None
                return type(self)(self.parent(), (x, self.value[1], self.value[2]))
            if i > 0:
                if self.value[0].epsilon(i) > self.value[2].phi(i):
                    x = self.value[0].e(i)
                    if x is None:
                        return None
                    return type(self)(self.parent(), (x, self.value[1], self.value[2]))
                else:
                    x = self.value[2].e(i)
                    if x is None:
                        return None
                    return type(self)(self.parent(), (self.value[0], self.value[1], x))
            # else i < 0
            M = self.parent()._cartan_type.m + 1
            if self.value[0].phi(i) < self.value[1].epsilon(M+i):
                x = self.value[1].e(M+i)
                if x is None:
                    return None
                return type(self)(self.parent(), (self.value[0], x, self.value[2]))
            else:
                x = self.value[0].e(i)
                if x is None:
                    return None
                return type(self)(self.parent(), (x, self.value[1], self.value[2]))

        def f(self, i):
            r"""
            Return the action of the crystal operator `f_i` on ``self``.

            EXAMPLES::

                sage: K = crystals.KacModule(['A', [2,2]], [2,1], [1])
                sage: mg = K.module_generator()
                sage: mg.f(-2)
                ({}, [[-3, -2], [-2]], [[1]])
                sage: mg.f(-1)
                ({}, [[-3, -3], [-1]], [[1]])
                sage: mg.f(0)
                ({-e[-1]+e[1]}, [[-3, -3], [-2]], [[1]])
                sage: mg.f(1)
                ({}, [[-3, -3], [-2]], [[2]])
                sage: mg.f(2)
                sage: b = mg.f_string([1,0,1,-1,-2,0,1,2,0,-2,-1,2,0]); b
                ({-e[-3]+e[3], -e[-2]+e[1], -e[-1]+e[1], -e[-1]+e[2]},
                 [[-3, -2], [-2]], [[3]])
            """
            if i == 0:
                x = self.value[0].f(i)
                if x is None:
                    return None
                return type(self)(self.parent(), (x, self.value[1], self.value[2]))
            if i > 0:
                if self.value[0].epsilon(i) < self.value[2].phi(i):
                    x = self.value[2].f(i)
                    if x is None:
                        return None
                    return type(self)(self.parent(), (self.value[0], self.value[1], x))
                else:
                    x = self.value[0].f(i)
                    if x is None:
                        return None
                    return type(self)(self.parent(), (x, self.value[1], self.value[2]))
            # else i < 0
            M = self.parent()._cartan_type.m + 1
            if self.value[0].phi(i) > self.value[1].epsilon(M+i):
                x = self.value[0].f(i)
                if x is None:
                    return None
                return type(self)(self.parent(), (x, self.value[1], self.value[2]))
            else:
                x = self.value[1].f(M+i)
                if x is None:
                    return None
                return type(self)(self.parent(), (self.value[0], x, self.value[2]))

        def weight(self):
            r"""
            Return weight of ``self``.

            EXAMPLES::

                sage: K = crystals.KacModule(['A', [3,2]], [2,1], [5,1])
                sage: mg = K.module_generator()
                sage: mg.weight()
                (2, 1, 0, 0, 5, 1, 0)
                sage: mg.weight().is_dominant()
                True
                sage: mg.f(0).weight()
                (2, 1, 0, -1, 6, 1, 0)
                sage: b = mg.f_string([2,1,-3,-2,-1,1,1,0,-2,-1,2,1,1,1,0,2,-3,-2,-1])
                sage: b.weight()
                (0, 0, 0, 1, 1, 4, 3)
            """
            e = self.parent().weight_lattice_realization().basis()
            M = self.parent()._cartan_type.m + 1
            wt = self.value[0].weight()
            wt += sum(c*e[i-M] for i,c in self.value[1].weight())
            wt += sum(c*e[i+1] for i,c in self.value[2].weight())
            return wt

#####################################################################
## Helper functions


def to_dual_tableau(elt):
    r"""
    Return a type `A_n` crystal tableau ``elt`` as a tableau expressed
    in terms of dual letters.

    The dual letter of `k` is expressed as `\overline{n+2-k}` represented
    as `-(n+2-k)`.

    EXAMPLES::

        sage: from sage.combinat.crystals.kac_modules import to_dual_tableau
        sage: T = crystals.Tableaux(['A',2], shape=[2,1])
        sage: ascii_art([to_dual_tableau(t) for t in T])
        [  -3 -3   -3 -2   -3 -1   -3 -1   -2 -1   -3 -3   -3 -2   -2 -2 ]
        [  -2   ,  -2   ,  -2   ,  -1   ,  -1   ,  -1   ,  -1   ,  -1    ]

    TESTS:

    Check that :issue:`23935` is fixed::

        sage: from sage.combinat.crystals.kac_modules import to_dual_tableau
        sage: T = crystals.Tableaux(['A',2], shape=[])
        sage: to_dual_tableau(T[0])
        []

        sage: Ktriv = crystals.KacModule(['A',[1,1]], [], [])
        sage: Ktriv.module_generator()
        ({}, [], [])
    """
    from sage.combinat.tableau import Tableau
    M = elt.parent().cartan_type().rank() + 2
    if not elt:
        return Tableau([])
    tab = [ [elt[0].value-M] ]
    for i in range(1, len(elt)):
        if elt[i-1] < elt[i] or (elt[i-1].value != 0 and elt[i-1] == elt[i]):
            tab.append([elt[i].value-M])
        else:
            tab[len(tab)-1].append(elt[i].value-M)
    for x in tab:
        x.reverse()
    return Tableau(tab).conjugate()


def latex_dual(elt):
    r"""
    Return a latex representation of a type `A_n` crystal tableau ``elt``
    expressed in terms of dual letters.

    The dual letter of `k` is expressed as `\overline{n+2-k}`.

    EXAMPLES::

        sage: from sage.combinat.crystals.kac_modules import latex_dual
        sage: T = crystals.Tableaux(['A',2], shape=[2,1])
        sage: print(latex_dual(T[0]))
        {\def\lr#1{\multicolumn{1}{|@{\hspace{.6ex}}c@{\hspace{.6ex}}|}{\raisebox{-.3ex}{$#1$}}}
        \raisebox{-.6ex}{$\begin{array}[b]{*{2}c}\cline{1-2}
        \lr{\overline{3}}&\lr{\overline{3}}\\\cline{1-2}
        \lr{\overline{2}}\\\cline{1-1}
        \end{array}$}
        }
    """
    M = elt.parent().cartan_type().rank() + 2
    from sage.combinat.tableau import Tableau
    from sage.combinat.output import tex_from_array
    # Modified version of to_tableau() to have the entries be letters
    #   rather than their values
    if not elt:
        return "{\\emptyset}"

    tab = [ ["\\overline{{{}}}".format(M-elt[0].value)] ]
    for i in range(1, len(elt)):
        if elt[i-1] < elt[i] or (elt[i-1].value != 0 and elt[i-1] == elt[i]):
            tab.append(["\\overline{{{}}}".format(M-elt[i].value)])
        else:
            l = len(tab)-1
            tab[l].append("\\overline{{{}}}".format(M-elt[i].value))
    for x in tab:
        x.reverse()

    T = Tableau(tab).conjugate()
    return tex_from_array([list(row) for row in T])



# sage_setup: distribution = sagemath-combinat
"""
Examples of combinatorial species
"""
# ****************************************************************************
#       Copyright (C) 2008 Mike Hansen <mhansen@gmail.com>,
#
#  Distributed under the terms of the GNU General Public License (GPL)
#
#    This code is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
#    General Public License for more details.
#
#  The full text of the GPL is available at:
#
#                  https://www.gnu.org/licenses/
# ****************************************************************************

from .set_species import SetSpecies
from .partition_species import PartitionSpecies
from .subset_species import SubsetSpecies
from .recursive_species import CombinatorialSpecies
from .characteristic_species import CharacteristicSpecies, SingletonSpecies, EmptySetSpecies
from .cycle_species import CycleSpecies
from .linear_order_species import LinearOrderSpecies
from .permutation_species import PermutationSpecies
from .empty_species import EmptySpecies
from .sum_species import SumSpecies
from .product_species import ProductSpecies
from .composition_species import CompositionSpecies
from .functorial_composition_species import FunctorialCompositionSpecies

from sage.misc.cachefunc import cached_function


@cached_function
def SimpleGraphSpecies():
    """
    Return the species of simple graphs.

    EXAMPLES::

        sage: S = species.SimpleGraphSpecies()
        sage: S.generating_series().counts(10)
        [1, 1, 2, 8, 64, 1024, 32768, 2097152, 268435456, 68719476736]
        sage: S.cycle_index_series()[:5]                                                # needs sage.libs.pari sage.modules
        [p[],
         p[1],
         p[1, 1] + p[2],
         4/3*p[1, 1, 1] + 2*p[2, 1] + 2/3*p[3],
         8/3*p[1, 1, 1, 1] + 4*p[2, 1, 1] + 2*p[2, 2] + 4/3*p[3, 1] + p[4]]
        sage: S.isotype_generating_series()[:6]                                         # needs sage.libs.pari sage.modules
        [1, 1, 2, 4, 11, 34]

    TESTS::

        sage: seq = S.isotype_generating_series().counts(6)[1:]                         # needs sage.modules
        sage: oeis(seq)[0]                              # optional - internet           # needs sage.modules
        A000088: ...

    ::

        sage: seq = S.generating_series().counts(10)[1:]                                # needs sage.modules
        sage: oeis(seq)[0]                              # optional - internet           # needs sage.modules
        A006125: a(n) = 2^(n*(n-1)/2).
    """
    E = SetSpecies()
    E2 = SetSpecies(size=2)
    WP = SubsetSpecies()
    P2 = E2 * E
    return WP.functorial_composition(P2)


@cached_function
def BinaryTreeSpecies():
    r"""
    Return the species of binary trees on `n` leaves.

    The species of binary trees `B` is defined by `B = X + B \cdot B`,
    where `X` is the singleton species.

    EXAMPLES::

        sage: B = species.BinaryTreeSpecies()
        sage: B.generating_series().counts(10)
        [0, 1, 2, 12, 120, 1680, 30240, 665280, 17297280, 518918400]
        sage: B.isotype_generating_series().counts(10)
        [0, 1, 1, 2, 5, 14, 42, 132, 429, 1430]
        sage: B._check()
        True

    ::

        sage: B = species.BinaryTreeSpecies()
        sage: a = B.structures([1,2,3,4,5])[187]; a
        2*((5*3)*(4*1))
        sage: a.automorphism_group()                                                    # needs sage.groups
        Permutation Group with generators [()]

    TESTS::

        sage: seq = B.isotype_generating_series().counts(10)[1:]
        sage: oeis(seq)[0]                              # optional -- internet
        A000108: Catalan numbers: ...
    """
    B = CombinatorialSpecies(min=1)
    X = SingletonSpecies()
    B.define(X + B * B)
    return B


@cached_function
def BinaryForestSpecies():
    """
    Return the species of binary forests.

    Binary forests are defined as sets of binary trees.

    EXAMPLES::

        sage: F = species.BinaryForestSpecies()
        sage: F.generating_series().counts(10)
        [1, 1, 3, 19, 193, 2721, 49171, 1084483, 28245729, 848456353]
        sage: F.isotype_generating_series().counts(10)                                  # needs sage.modules
        [1, 1, 2, 4, 10, 26, 77, 235, 758, 2504]
        sage: F.cycle_index_series()[:7]                                                # needs sage.modules
        [p[],
         p[1],
         3/2*p[1, 1] + 1/2*p[2],
         19/6*p[1, 1, 1] + 1/2*p[2, 1] + 1/3*p[3],
         193/24*p[1, 1, 1, 1] + 3/4*p[2, 1, 1] + 5/8*p[2, 2] + 1/3*p[3, 1] + 1/4*p[4],
         907/40*p[1, 1, 1, 1, 1] + 19/12*p[2, 1, 1, 1] + 5/8*p[2, 2, 1] + 1/2*p[3, 1, 1] + 1/6*p[3, 2] + 1/4*p[4, 1] + 1/5*p[5],
         49171/720*p[1, 1, 1, 1, 1, 1] + 193/48*p[2, 1, 1, 1, 1] + 15/16*p[2, 2, 1, 1] + 61/48*p[2, 2, 2] + 19/18*p[3, 1, 1, 1] + 1/6*p[3, 2, 1] + 7/18*p[3, 3] + 3/8*p[4, 1, 1] + 1/8*p[4, 2] + 1/5*p[5, 1] + 1/6*p[6]]

    TESTS::

        sage: seq = F.isotype_generating_series().counts(10)[1:]                        # needs sage.modules
        sage: oeis(seq)[0]                              # optional - internet           # needs sage.modules
        A052854: Number of forests of ordered trees on n total nodes.
    """
    B = BinaryTreeSpecies()
    S = SetSpecies()
    F = S(B)
    return F


del cached_function  # so it doesn't get picked up by tab completion



# sage_setup: distribution = sagemath-combinat
from sage.structure.sage_object cimport SageObject

cdef class RiggedPartition(SageObject):
    cdef public list _list
    cdef public list vacancy_numbers
    cdef public list rigging
    cdef long _hash

    cpdef get_num_cells_to_column(self, int end_column, t=*)
    cpdef insert_cell(self, int max_width)
    cpdef remove_cell(self, row, int num_cells=*)

cdef class RiggedPartitionTypeB(RiggedPartition):
    pass
